/* -*- c++ -*-
 * tkresolver
 *
 * Copyright (C) 2004-2007 Robert Ham <rah@bash.sh>
 *
 * GPL
 *
 */

#ifndef __TK_STRING_H__
#define __TK_STRING_H__

#include <wchar.h>

#include <string>
#include <sstream>

#include <glibmm.h>
#include <gmpxx.h>

#include "mbs.h"

namespace TkResolver {

  class String : public std::wstring
    {
    public:
      String() {}
      String(const wchar_t * wstr)
	: std::wstring(wstr) {}
      String(const char * str)
	: std::wstring(mbs_to_wstring(str)) {}
      String(const std::wstring & wstr)
	: std::wstring(wstr) {}
      String(const std::string & sstr)
	: std::wstring(string_to_wstring(sstr)) {}
      String(const Glib::ustring & ustr)
	: std::wstring(ustring_to_wstring(ustr)) {} 
      
      String(bool b)               { std::wostringstream o; o << b; *this = o.str(); }
      String(char c)               { std::ostringstream  o; o << c; *this = o.str(); }
      String(int i)                { std::wostringstream o; o << i; *this = o.str(); }
      String(unsigned i)           { std::wostringstream o; o << i; *this = o.str(); }
      String(long i)               { std::wostringstream o; o << i; *this = o.str(); }
      String(unsigned long i)      { std::wostringstream o; o << i; *this = o.str(); }
      String(long long i)          { std::wostringstream o; o << i; *this = o.str(); }
      String(unsigned long long i) { std::wostringstream o; o << i; *this = o.str(); }
      String(float f)              { std::wostringstream o; o << f; *this = o.str(); }
      String(double d)             { std::wostringstream o; o << d; *this = o.str(); }
      String(long double d)        { std::wostringstream o; o << d; *this = o.str(); }
      
      String(const mpz_class &z)
	{ std::ostringstream o; o << z; *this = o.str(); }
      String(const mpf_class &f, unsigned long precision = 100)
	{ std::ostringstream o; o.precision(precision); o << f; *this = o.str(); }
      
      /** Stringify a rational into the form
       *
       *    1.1111111111111 (222222222222222222222222222222222222/33333333333...)
       */
      String(const mpq_class &q, unsigned long precision = 30, size_type max_len = 100);
      
      String & operator= (const wchar_t * wstr)
	{ return *this = String(wstr); }
      String & operator= (const char * str)
	{ return *this = String(str); }
      String & operator= (const std::string & sstr)
	{ return *this = string_to_wstring(sstr); }
      String & operator= (const Glib::ustring & ustr)
        { return *this = ustring_to_wstring(ustr); }
      
      operator std::string() const
	{ return wstring_to_string(*this); }
      operator Glib::ustring() const
        { return wstring_to_ustring(*this); }
    };
  
  class StringList : public std::list<String>
    {
    public:
      StringList() {}
      template <class InputIterator>
      StringList(InputIterator f, InputIterator l)
	: std::list<String>(f, l) {}
      StringList(iterator &start, iterator &end)
	: std::list<String>(start, end) {}
      StringList merge(const StringList &list) const;
      StringList uniq() const;
    };

}; /* namespace TkResolver */


/* Global operators */

/* + */
/*static TkResolver::String operator+(const TkResolver::String &tkstr1, const TkResolver::String &tkstr2)
{ return static_cast<std::wstring>(tkstr1) + static_cast<std::wstring>(tkstr2); }
static std::string operator+(const std::string &str, const TkResolver::String &tkstr)
{ return str + static_cast<std::string>(tkstr); }
static TkResolver::String operator+(const TkResolver::String &tkstr, const std::string &str)
{ return tkstr + static_cast<TkResolver::String>(str); }
static TkResolver::String operator+(const TkResolver::String &tkstr, const char *str)
{ return tkstr + static_cast<TkResolver::String>(str); }*/

/* += */
/*static TkResolver::String operator+=(const TkResolver::String &tkstr1, const TkResolver::String &tkstr2)
{ return static_cast<std::wstring>(tkstr1) += static_cast<std::wstring>(tkstr2); }
static TkResolver::String operator+=(const TkResolver::String &tkstr, const std::string &str)
{ return tkstr += static_cast<TkResolver::String>(str); }
static TkResolver::String operator+=(const TkResolver::String &tkstr, const char *str)
{ return tkstr += static_cast<TkResolver::String>(str); }*/

/* == */
static bool operator==(const TkResolver::String &tkstr1, const TkResolver::String &tkstr2)
{ return static_cast<std::wstring>(tkstr1) == static_cast<std::wstring>(tkstr2); }
static bool operator==(const std::string &str, const TkResolver::String &tkstr)
{ return static_cast<TkResolver::String>(str) == tkstr; }
static bool operator==(const char *str, const TkResolver::String &tkstr)
{ return static_cast<TkResolver::String>(str) == tkstr; }

#endif /* __TK_STRING_H__ */
