/* -*- c++ -*-
 * tkresolver - Telekinesis testing software
 *
 * Copyright (C) 2004-2007 Robert Ham <rah@bash.sh>
 *
 * GPL all the way, baby :)
 *
 */

template<typename S>
StringList
display_column_names(const std::list< DisplayColumnInfo<S> > & display_columns)
{
  StringList columns;
  bool exists;
  for (typename std::list< DisplayColumnInfo<S> >::const_iterator iter(display_columns.begin());
       iter != display_columns.end();
       ++iter)
    {
      for (StringList::const_iterator dciter(iter->columns.begin());
	   dciter != iter->columns.end();
	   ++dciter)
	{
	  columns.push_back(*dciter);
	}
    }

  return columns;
}

template<typename V, typename CV>
static V
iter_get_column_value(Gtk::TreeModel::iterator & iter, unsigned column_index)
{
  CV value;
  iter->get_value(column_index, value);
  return static_cast<V>(value);
}

template<>
static Db::Id
iter_get_column_value<Db::Id, Glib::ustring>(Gtk::TreeModel::iterator & iter, unsigned column_index)
{
  Glib::ustring value;
  iter->get_value(column_index, value);
  
  std::istringstream istr(value);
  Db::Id id;
  istr >> id;
  
  return id;
}
// FIXME: Bad and Wrong
template<> static Db::Id
iter_get_column_value<Db::Id, Db::Date>(Gtk::TreeModel::iterator & iter, unsigned column_index)
{ throw std::logic_error("attempt to retrieve Db::Id from Db::Date"); }
template<> static Db::Id
iter_get_column_value<Db::Id, Db::DateTime>(Gtk::TreeModel::iterator & iter, unsigned column_index)
{ throw std::logic_error("attempt to retrieve Db::Id from Db::DateTime"); }
template<> static String
iter_get_column_value<String, Db::Id>(Gtk::TreeModel::iterator & iter, unsigned column_index)
{ throw std::logic_error("attempt to retrieve String from Db::Id"); }
template<> static String
iter_get_column_value<String, long double>(Gtk::TreeModel::iterator & iter, unsigned column_index)
{ throw std::logic_error("attempt to retrieve String from long double"); }
template<> static String
iter_get_column_value<String, double>(Gtk::TreeModel::iterator & iter, unsigned column_index)
{ throw std::logic_error("attempt to retrieve String from double"); }
template<> static String
iter_get_column_value<String, float>(Gtk::TreeModel::iterator & iter, unsigned column_index)
{ throw std::logic_error("attempt to retrieve String from float"); }
template<> static String
iter_get_column_value<String, Db::Date>(Gtk::TreeModel::iterator & iter, unsigned column_index)
{ throw std::logic_error("attempt to retrieve String from Db::Date"); }
template<> static String
iter_get_column_value<String, Db::DateTime>(Gtk::TreeModel::iterator & iter, unsigned column_index)
{ throw std::logic_error("attempt to retrieve String from Db::DateTime"); }
template<> static std::string
iter_get_column_value<std::string, char>(Gtk::TreeModel::iterator & iter, unsigned column_index)
{ throw std::logic_error("attempt to retrieve std::string from char"); }
template<> static std::string
iter_get_column_value<std::string, bool>(Gtk::TreeModel::iterator & iter, unsigned column_index)
{ throw std::logic_error("attempt to retrieve std::string from bool"); }
template<> static std::string
iter_get_column_value<std::string, int>(Gtk::TreeModel::iterator & iter, unsigned column_index)
{ throw std::logic_error("attempt to retrieve std::string from int"); }
template<> static std::string
iter_get_column_value<std::string, short>(Gtk::TreeModel::iterator & iter, unsigned column_index)
{ throw std::logic_error("attempt to retrieve std::string from short"); }
template<> static std::string
iter_get_column_value<std::string, Db::Id>(Gtk::TreeModel::iterator & iter, unsigned column_index)
{ throw std::logic_error("attempt to retrieve std::string from Db::Id"); }
template<> static std::string
iter_get_column_value<std::string, long double>(Gtk::TreeModel::iterator & iter, unsigned column_index)
{ throw std::logic_error("attempt to retrieve std::string from long double"); }
template<> static std::string
iter_get_column_value<std::string, double>(Gtk::TreeModel::iterator & iter, unsigned column_index)
{ throw std::logic_error("attempt to retrieve std::string from double"); }
template<> static std::string
iter_get_column_value<std::string, float>(Gtk::TreeModel::iterator & iter, unsigned column_index)
{ throw std::logic_error("attempt to retrieve std::string from float"); }
template<> static std::string
iter_get_column_value<std::string, Db::Date>(Gtk::TreeModel::iterator & iter, unsigned column_index)
{ throw std::logic_error("attempt to retrieve std::string from Db::Date"); }
template<> static std::string
iter_get_column_value<std::string, Db::DateTime>(Gtk::TreeModel::iterator & iter, unsigned column_index)
{ throw std::logic_error("attempt to retrieve std::string from Db::DateTime"); }

template<typename V>
V
DataSelector:: get_column_value(Gtk::TreeModel::iterator & iter, const String &column_name) const
{
  Db::Type type(  get_column_type(column_name)  );
  unsigned index( get_column_index(column_name) );
  
  using namespace odbc;
  switch(type)
    {
    case Types::VARCHAR:
    case Types::LONGVARCHAR:
      return iter_get_column_value<V, Glib::ustring>(iter, index);
    case Types::CHAR:
      return iter_get_column_value<V, char>(iter, index);
    case Types::BIT:
      return iter_get_column_value<V, bool>(iter, index);
    case Types::INTEGER:
      return iter_get_column_value<V, int>(iter, index);
    case Types::BIGINT:
      return iter_get_column_value<V, Db::Id>(iter, index);
    case Types::SMALLINT:
    case Types::TINYINT:
      return iter_get_column_value<V, short>(iter, index);
    case Types::DECIMAL:
    case Types::NUMERIC:
      return iter_get_column_value<V, long double>(iter, index);
    case Types::DOUBLE:
    case Types::REAL:
      return iter_get_column_value<V, double>(iter, index);
    case Types::FLOAT:
      return iter_get_column_value<V, float>(iter, index);
    case Types::DATE:
      return iter_get_column_value<V, Db::Date>(iter, index);
    case Types::TIMESTAMP:
      return iter_get_column_value<V, Db::DateTime>(iter, index);
    default:
      throw std::runtime_error(std::string("unknown data type (") + static_cast<std::string>(String(type)) + ") for data column");
    }
}

template<typename SV>
Selection<SV>
DataSelector:: get_selection() const
{
  Selection<SV> sel;
  ModelIteratorList selections(get_selected());
  
  for (ModelIteratorList::iterator sel_iter(selections.begin());
       sel_iter != selections.end();
       ++sel_iter)
    {
      SelectionValues<SV> values;
      for (StringList::const_iterator col_iter(_selection_columns.begin());
	   col_iter != _selection_columns.end();
	   ++col_iter)
	values.insert( typename SelectionValues<SV>::value_type(*col_iter, get_column_value<SV>(*sel_iter, *col_iter)) );

      sel.push_back(values);
    }
  
  return sel;
}

template<typename SV>
void
DataSelector:: set_selection(const String &column_name, const SV &value) {
  for (Gtk::TreeModel::iterator iter( _store->get_iter("0") );
       iter;
       ++iter)
    {
      if (get_column_value<SV>(iter, column_name) == value) {
	select(iter);
	break;
      }
    }
}

template<typename C>
Gtk::CellRenderer *
DataSelector:: create_view_cell(const DisplayColumnInfo<typename C::SlotCellData> & display_column, DisplayColumnAdaptor<C> & adaptor)
{
  Gtk::CellRenderer * cell;
  if (display_column.slot != 0) /* have slot */
    {
      WDEBUG_OUT << "callback column '" << display_column.title << "'" << std::endl;

      cell = Gtk::manage(new Gtk::CellRendererText);
      adaptor.pack_start(*cell);
      adaptor.set_cell_data_func(*cell, *display_column.slot);
    }
  else /* don't have slot; just pack renderers into column */
    {
      WDEBUG_OUT << "packing cells for column '" << display_column.title << "'" << std::endl;

      for (StringList::const_iterator iter( display_column.columns.begin() );
	   iter != display_column.columns.end();
	   ++iter)
	{
	  WDEBUG_OUT << "packing cell '" << *iter << "' for column '" << display_column.title << "'" << std::endl;

	  cell = create_cell(*iter);
	  adaptor.pack_start(*cell);
	  adaptor.add_attribute(*cell, get_column_view_attribute(*iter), get_column_index(*iter));
	}
    }

  return cell;
}



template<typename C>
DisplayColumnAdaptor<C>:: DisplayColumnAdaptor(C * column)
  : _column(column)
{
}

template<typename C>
void
DisplayColumnAdaptor<C>:: pack_start(Gtk::CellRenderer & cell)
{
  _column->pack_start(cell);
}

template<typename C>
void
DisplayColumnAdaptor<C>:: set_cell_data_func(Gtk::CellRenderer & cell, typename C::SlotCellData & slot)
{
  _column->set_cell_data_func(cell, slot);
}

template<typename C>
void
DisplayColumnAdaptor<C>:: add_attribute(Gtk::CellRenderer & cell, const Glib::ustring & attribute, int column_index)
{
  _column->add_attribute(cell, attribute, column_index);
}
